from abc import ABC, abstractmethod
from typing import Any, Dict, Optional, Tuple, Union

from FaceEngine import Detection  # pylint: disable=E0611,E0401

from .image_utils.geometry import LANDMARKS, Point, Rect


class BaseEstimation(ABC):
    """
    Base class for estimation structures.

    Attributes:
        _coreEstimation: core estimation
    """

    __slots__ = ("_coreEstimation",)

    def __init__(self, coreEstimation: Any):
        self._coreEstimation = coreEstimation

    @property
    def coreEstimation(self):
        """
        Get core estimation from init
        Returns:
            _coreEstimation
        """
        return self._coreEstimation

    @abstractmethod
    def asDict(self) -> Union[dict, list]:
        """
        Convert to a dict.

        Returns:
            dict from luna api
        """
        pass

    def __repr__(self) -> str:
        """
        Representation.

        Returns:
            str(self.asDict())
        """
        return str(self.asDict())


class Landmarks(BaseEstimation):
    """
    Base class for landmarks

    Attributes:
        _points (Optional[Tuple[Point[float]]]): lazy loaded attributes: core landmarks as point list
    """

    __slots__ = ["_points", "_coreEstimation"]

    def __init__(self, coreLandmarks: LANDMARKS):
        """
        Init

        Args:
            coreLandmarks (LANDMARKS): core landmarks
        """
        super().__init__(coreLandmarks)
        self._points: Optional[Tuple[Point[float], ...]] = None

    @property
    def points(self) -> Tuple[Point[float], ...]:
        """
        Lazy points loader.

        Returns:
            list of points
        """
        if self._points is None:
            self._points = tuple(
                Point.fromVector2(self._coreEstimation[index]) for index in range(len(self._coreEstimation))
            )
        return self._points

    def asDict(self) -> Tuple[Tuple[int, int], ...]:  # type: ignore
        """
        Convert to dict

        Returns:
            list to list points
        """
        pointCount = len(self._coreEstimation)
        points = self._coreEstimation
        return tuple(((int(points[index].x), int(points[index].y)) for index in range(pointCount)))


class BoundingBox(BaseEstimation):
    """
    Detection bounding box, it is characterized of rect and score:

        - rect (Rect[float]): face bounding box
        - score (float): face score (0,1), detection score is the measure of classification confidence
                         and not the source image quality. It may be used topick the most "*confident*" face of many.
    """

    #  pylint: disable=W0235
    def __init__(self, boundingBox: Detection):
        """
        Init.

        Args:
            boundingBox: core bounding box
        """
        super().__init__(boundingBox)

    @property
    def score(self) -> float:
        """
        Get score

        Returns:
            number in range [0,1]
        """
        return self._coreEstimation.getScore()

    @property
    def rect(self) -> Rect[float]:
        """
        Get rect.

        Returns:
            float rect
        """
        return Rect.fromCoreRect(self._coreEstimation.getRect())

    def asDict(self) -> Dict[str, Union[Dict[str, float], float]]:
        """
        Convert to  dict.

        Returns:
            {"rect": self.rect, "score": self.score}
        """
        return {"rect": self.rect.asDict(), "score": self.score}
